// Motor Controller Main Board

// Pin Definitions
#DEFINE  TX       PIN_C6
#DEFINE  RX       PIN_C7
#DEFINE  MSDA     PIN_C4
#DEFINE  MSCL     PIN_C5
#DEFINE  GO       PIN_C3
#DEFINE  LED_0    PIN_A6
#DEFINE  LED_1    PIN_C0
#DEFINE  LED_2    PIN_C1
#DEFINE  LED_3    PIN_C2
#DEFINE  ID0      PIN_A0
#DEFINE  ID1      PIN_A1

#DEFINE  QIX      PIN_A2
#DEFINE  QEA      PIN_A3
#DEFINE  QEB      PIN_A4
#DEFINE  PWM      PIN_B0
#DEFINE  MOTOREN  PIN_B5
#DEFINE  MOTOREN2 PIN_B4

//note the different PIC used
#include <18f2431.h>

//#device ICD=TRUE
#fuses EC_IO, NOLVP, NOWDT, NOPROTECT
#device high_ints=TRUE
#DEVICE ADC=8

#include <math.h>
#use delay(clock=40000000)     
//#use i2c(SLAVE, FAST, SCL=MSCL, SDA=MSDA, address=0x80, FORCE_HW)   // Hardware SSP Module
#use i2c(SLAVE, SLOW=50000, SCL=MSCL, SDA=MSDA, address=0x80, FORCE_HW)   // Hardware SSP Module

// 20 KHz pwm freq with 40 MHz osc.
#DEFINE POWER_PWM_PERIOD 499

// Register Definitions
// SSP Module Control Register
#byte    SSPCON   = 0xFC6
// QEI Module Control Register
#byte    QEICON   = 0xFB6
#bit     MYDIR    = QEICON.5
// QEI Rollover INT bit
#byte    PIR3     = 0xFA4
#bit     MYQEI    = PIR3.2
// POSCNT Register
volatile int16    POSCNT   = 0x0000;
#byte    POSCNT   = 0xF66  
#byte    POSCNTH  = 0xF67
#byte    POSCNTL  = 0xF66
// MAXCNT Register
int16    MAXCNT   = 0xFFFF;
#byte    MAXCNT   = 0xF64  
#byte    MAXCNTH  = 0xF65
#byte    MAXCNTL  = 0xF64
// DLFTCON
int8 DLFTCON      = 0x31;//0b00110001;
#byte DLFTCON     = 0xF60

// Safety
volatile int1           needToTrapezoidalCompute = 0;


// Global Variables
int8           trap_track = 1;
int8           slaveID;
signed int16   duty;
int16          an0_r1;
int16          an1_r1;
int8           go_count = 0;

//PID vars
volatile int32          TotalCount      = 0x80000000;
int32                   TotalCountStore = 0x80000000;
volatile int16          TotalCountH     = 0x8000;
signed int32   p_error     = 0;

// Trapezoidal Tracking Variables
int32          TargetPosition = 0x80000000;
signed int32   vel_max     = 1;
signed int32   sudoVmax    = 1;
signed int32   acc_desired = 0;
signed int32   vel_desired = 0;
int32          pos_desired = 0x80000000;
int32          actual_time = 0;
int32          t1          = 0;
int32          t2          = 0;
int32          t3          = 0;
int32          d1, d2, d3;
signed int16   accel_status = 0;
signed int16   accel_postscaler = 0;
signed int     directionbit = 1;

// I2C Variables
int8           state;                              // I2C state
int32          Number32Bit = 0;                    // Stores numbers converted to 32 bit
int8           Number[4]   = {0,0,0,0};            // Stores 4 bytes of 32 bit numbers
int8           DataType    = 0;                    // Type of operation being performed

// Function Prototypes
void setup();
void trapezoidalCompute();
void trapezoidalTrack();
void PID();
void updateStatus();
void ssp_interrupt();
void processI2C();
void motorgo();
void qei_rollover_int();
void flashLEDs();
void low_volt_detect();
long uisqrt32(int32 r);

// Set up the Hardware Modules on PIC and initialize variables
void setup()
{
   // Disable Interrupts
   disable_interrupts(GLOBAL);
  
   // Disable H-bridge
   output_low(MOTOREN);
  
   // Setup PWM Ports (PWM0 + PWM1 @ 20Khz)
   setup_power_pwm_pins(PWM_BOTH_ON, PWM_OFF, PWM_OFF, PWM_OFF);
  
   // Setup PWM Mode
   setup_power_pwm(PWM_FREE_RUN, 1, 0, POWER_PWM_PERIOD, 0, 1,0);

   // Set PWM Duty to 50%
   duty = 1000;
   set_power_pwm0_duty((int16) duty);

   // Setup ADC
   setup_adc_ports(sAN0);
   setup_adc_ports(sAN1);
   setup_adc(ADC_CLOCK_INTERNAL);

   // Determine Module ID #
   set_adc_channel(0);
   delay_us(10);
   an0_r1 = read_adc();
   set_adc_channel(1);
   delay_us(10);
   an1_r1 = read_adc();
   if ( an0_r1 > an1_r1)
   {
      slaveID = 0x80;
   }
   else
   {
      slaveID = 0x82;
   }
  
   // Reset I2C Slave Address according to pins
   I2C_SlaveAddr(slaveID);
  
   // I2C Hardware Interrupt
   enable_interrupts(INT_SSP);
  
   // EXT INT (GO PIN)
   enable_interrupts(INT_EXT);
  
   // INT_LVD
   enable_interrupts(INT_LOWVOLT);
   
   // Timer 5
   //setup_timer_5(T5_INTERNAL|T5_DIV_BY_1);
   setup_timer_1(T1_INTERNAL|T1_DIV_BY_1);  
  
   // QEI Control Register
   // No VELMODE, INTERRUPT ON ROLLOVER, DIR=UP, 4x MODE, VEL/64
   QEICON = 0xFB;
   //QEICON = 0xEB;
  
   // Setup MAXCNT
   MAXCNT = 0xFFFF;
  
   // Setup POSCNT
   POSCNT = 0x0000;
  
   // Digital Filters
   DLFTCON = 0x31;
  
   // Reset Initial QEI Interrupt Vector
   MYQEI=0;
   // Enable QEI Rollover Interrupt
   enable_interrupts(INT_IC2QEI);
  
   // Setup Initial Motor Position
   TotalCount = 0x80000000;
  
   accel_postscaler = 35;
}

// Pre-process the trapezoidal profile parameters
void trapezoidalCompute()
{
   p_error = (signed int32) (pos_desired - TotalCount);
   
   acc_desired = 1;
   directionbit = 1;
   vel_max = abs(vel_max);

   t1 = vel_max * (accel_postscaler+1);
   d1 = t1 * t1;
   d1 /= (2*(accel_postscaler+1));
   
   d2 = abs(abs(p_error) - d1 - d1);
   t2 = (d2 / vel_max) + t1;
   d2 += d1;

   t3 = t2 + t1;
   d3 = d1 + d2;

   if (p_error < 0)
   {
      vel_max = -vel_max;
      
      acc_desired = -1;
      directionbit = -1;
   }
   sudoVmax = vel_max;

   while(abs(d1+d1)>abs(p_error))
   {
      sudoVmax = abs(sudoVmax);
      sudoVmax--;
 
      t1 = sudoVmax *(accel_postscaler+1);
      d1 = t1 * t1;
      d1 /= (2*(accel_postscaler+1));
     
      d2 = abs(abs(p_error) - d1 - d1);
      t2 = (d2 / sudoVmax) + t1;
      d2 += d1;
   
      t3 = t2 + t1;
      d3 = d1 + d2;

      if (p_error < 0)
      {
         sudoVmax = -sudoVmax;
         acc_desired = -1;
         directionbit = -1;
      }
   }
}



// Perform the trapezoidal increasing of position using Euler approximation
void trapezoidalTrack()                               // Trapezoidal Position Tracking with vel/acc control
{
   if (trap_track == 1)
   {
      if (actual_time < t1)                              // Phase 1, position increasing, velocity increasing
      {
         if(accel_status<accel_postscaler)
         {
            acc_desired = 0;
            accel_status++;
         }
         else
         {
            acc_desired = directionbit;
            accel_status = 0;
         }
         vel_desired += acc_desired;
         TargetPosition += vel_desired;
      }
      else if (actual_time < t2)                         // Phase 2, position increasing, velocity constant
      {
         vel_desired = sudoVmax;
         TargetPosition += vel_desired;
         accel_status = 0;
      }
      else if (actual_time < t3)                         // Phase 3, position increasing, velocity decreasing
      {
         if(accel_status<accel_postscaler)
         {
            acc_desired = 0;
            accel_status++;
         }
         else
         {
            acc_desired = directionbit;
            accel_status = 0;
         }
         vel_desired -= acc_desired;
         TargetPosition += vel_desired;
      }
      else                                               // Phase 4, position constant, velocity 0
      {
         TargetPosition = pos_desired;
         vel_desired = 0;
      }
     
      // Increase the time variable
      actual_time++;
      // PID the new position
      PID();
   }
}

// Proportional position control
void PID()                                           
{
   updateStatus();
   p_error = (signed int32) ((signed int32) TargetPosition - (signed int32) TotalCount);             // Compute proportional error
   duty = 1000;                                       // Set Duty to 50%
   duty += p_error;                                   // Adjust Duty of motor (wrt to error) (k=1)
  
   if (duty > 1990)                                   // Check for Saturation
   {
      duty = 1990;
   }
   else if (duty < 10)
   {
      duty = 10;
   }

   
   disable_interrupts(GLOBAL);
   set_power_pwm0_duty((int16)duty);                  // Adjust PWM in response to PID control
   enable_interrupts(GLOBAL);
}

void updateStatus()                   
{
   TotalCount = make32((int16)TotalCountH, (int16)POSCNT);

}

// Quadrature Encoder Module Rollover Interrupt
#INT_IC2QEI HIGH
void qei_rollover_int()
{
   updateStatus();
   TotalCountStore = TotalCount;

   if (MYDIR == 1)                                    // Adjust High Bytes of TotalCount
   {
      TotalCountH++; 
   }
   else
   {
      TotalCountH--;
   }

   updateStatus();
   if (((int32) TotalCount) > ((int32) 0x00018000 + (int32) TotalCountStore))
   {
      TotalCountH--;
   }
   else if (((int32) TotalCountStore) > ((int32) TotalCount + (int32) 0x00018000))
   {
      TotalCountH++;
   }
}

void main()
{
   // Setup PIC
   setup();
  
   // Enable Global Interrupts
   enable_interrupts(GLOBAL);
  
   // Enable H-Bridge
   output_high(MOTOREN);
   output_high(MOTOREN2);

   pos_desired = 0x80000000;
  
   // Main Motor Control Loop
   while (TRUE)
   {
      // Compute 
      if (needToTrapezoidalCompute == 1)
      {
         set_power_pwm0_duty((int16)1000);
         actual_time = 0;
         trap_track = 0;
         pos_desired = pos_desired + (signed int32)Number32Bit;
         trapezoidalCompute();
         actual_time = 0;
         trap_track = 1;
         needToTrapezoidalCompute = 0;
      }
      // Interrupt every 2.5 Khz (400us)
      else if (get_timer1()>4000)
      {
         set_timer1(0);
         trapezoidalTrack();                                // Track with Trapezoidal Trakcing
      }
   }

}


// I2C Hardware Interrupt
#INT_SSP high
void ssp_interrupt()
{
   disable_interrupts(global);                        // Disable interrupts necessary to prevent exit during transmission
   state = i2c_isr_state();
   if(state < 0x80)                                   // Master is sending data
   {
      if(state == 0)                                  // First received byte is address
      {
      }
      if(state == 1)                                  // Second received byte is address (Data packet)
      {
         DataType = i2c_read();
      }
      if(state > 1)                                    // Additional received bytes are data
      {
         Number[state-2] = i2c_read();
      }
      if (state == 5)                                   // Combine 4 one byte numbers into one 32 bit number
      {
         Number32Bit = make32(Number[3], Number[2], Number[1], Number[0]);
      }
   }
   else if(state >= 0x80)                                  //Master is requesting data
   {
      updateStatus();
      switch(DataType)                                //Slices data to send to master
      {
         case 255:
            i2c_write((int8)(TotalCount));
            break;
         case 254:
            i2c_write((int8)(TotalCount>>8));
            break;
         case 253:
            i2c_write((int8)(TotalCount>>16));
            break;
         case 252:
            i2c_write((int8)(TotalCount>>24));
            break;
         case 250:
            if ((actual_time < t3) || (abs(p_error) > 500) || (needToTrapezoidalCompute == 1))
            {   
               // NOT READY
               i2c_write((int8) 0);
            }
            else
            {   
               // READY
               i2c_write((int8) 1);
            }
            break;
      }
   }
   enable_interrupts(global);                         // Re-enable interrupts
  
}

void processI2C()                                     // Updates variables on external signal on line B0
{
   switch(DataType)
   {
      case 1:                                         // Mast command d_output
         break;
      case 2:                                         // Target position update (Relative)
         needToTrapezoidalCompute = 1;
         break;
      case 3:                                         // Target velocity update
         vel_max = (signed int32)Number32Bit;
         break;
      case 4:                                         // Reset and Abort
         TotalCount     = 0x80000000;
         TotalCountH    = 0x8000;
         TargetPosition = 0x80000000;
         pos_desired    = 0x80000000;
        
         vel_max = 0;
         acc_desired = 1;
         t1 = 0;
         t2 = 0;
         t3 = 0;
         actual_time = 0;
         break;
      case 5:                                         // Abort ONLY
         updateStatus();
         TargetPosition = TotalCount;
         pos_desired    = TotalCount;
         t1 = 0;
         t2 = 0;
         t3 = 0;
         break;
      case 6: 
         // Crazy Abort
         updateStatus();
         if (actual_time < t3)
         {
            if (directionbit == -1)
            {
               pos_desired  = TotalCount - d1;
            }
            else
            {
               pos_desired  = TotalCount + d1;
            }
            vel_desired = sudoVmax;
            accel_status = 0;
            t1 = 0;
            t2 = 0;
            t3 = actual_time;
         }
         break;
      case 7:
         accel_postscaler = Number32Bit;
         break;
         
      case 10:                                        // Reset position variable
         POSCNT         = 0x0000;
         TotalCountH    = 0x8000;
         TotalCount     = 0x80000000;
         TargetPosition = 0x80000000;
         pos_desired    = 0x80000000;
        
         vel_max = 0;
         acc_desired = 0;
         vel_desired = 0;

         t1 = 0;
         t2 = 0;
         t3 = 0;
        
         break;
   }
   DataType = 99;                                     // Prevent from double computation
}

// EXT INT GO PIN
#INT_EXT high     
void motorgo()
{
   go_count++;
   processI2C();
}

#INT_LOWVOLT
void low_volt_detect()
{
   disable_interrupts(GLOBAL);
   while(true)
   {
   }
}
